// PC-8001 emulation CGI
// Copyright (C) 2015 by Yasuo Kuwahara

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/time.h>
#include "Z80.h"

#define StartBASIC()	(m[0xeb54] | m[0xeb55] << 8)
#define EndBASIC()		(m[0xefa0] | m[0xefa1] << 8)
#define KEYIN_HOOK		0xf75
#define LOAD_HOOK		0x3cec
#define EXTENDER_CODE	0xe7
#define ex16(x)			((x) << 8 & 0xff00 | (x) >> 8 & 0xff)

struct Key {
	Key() { for (int i = 0; i < 16; i++) d[i] = 255; }
	void Changed(int v, int f) {
		int port = v >> 3 & 0xf;
		int mask = 1 << (v & 7);
		if (f) d[port] |= mask;
		else d[port] &= ~mask;
	}
	uint8_t d[16];
};

struct RealTimeClock {
	RealTimeClock() : data(0), pos(0), tofs(0) {
		for (int i = 0; i < 5; i++) t[i] = tmp[i] = 0;
	}
	void Strobe(int v) {
		time_t c;
		tm tm;
		switch (v & 0xf) {
			case 1:
				pos = 0;
				for (int i = 0; i < 5; i++) tmp[i] = 0;
				c = time(NULL) - tofs;
				localtime_r(&c, &tm);
				t[4] = (tm.tm_mon + 1) << 4;
				t[3] = (tm.tm_mday / 10 << 4) + tm.tm_mday % 10;
				t[2] = (tm.tm_hour / 10 << 4) + tm.tm_hour % 10;
				t[1] = (tm.tm_min / 10 << 4) + tm.tm_min % 10;
				t[0] = (tm.tm_sec / 10 << 4) + tm.tm_sec % 10;
				data = t[0] & 1;
				break;
			case 2:
				for (int i = 0; i < 5; i++) t[i] = tmp[i];
				tm.tm_year = 79;
				tm.tm_mon = (tmp[4] >> 4) - 1;
				tm.tm_mday = (tmp[3] >> 4) * 10 + (tmp[3] & 15);
				tm.tm_hour = (tmp[2] >> 4) * 10 + (tmp[2] & 15);
				tm.tm_min = (tmp[1] >> 4) * 10 + (tmp[1] & 15);
				tm.tm_sec = (tmp[0] >> 4) * 10 + (tmp[0] & 15);
				tofs = time(NULL) - mktime(&tm) - 3600;
				break;
		}
	}
	void Shift(int v) {
		if (pos < 40) {
			tmp[pos >> 3] |= (v & 1) << (pos - 3 & 7);
			pos++;
			data = t[pos >> 3] >> (pos & 7) & 1;
		}
	}
	int data, pos;
	time_t tofs;
	int t[5], tmp[5];
};

struct CRTController {
	CRTController() : seq(0), x(-1), y(-1), width80(false), height25(false), colorMode(false) {}
	void Command(int data) {
		switch (data) {
			case 0:
				seq = 5;
				break;
			case 0x80:
				x = y = -1;
				break;
			case 0x81:
				seq = 7;
				break;
		}
	}
	void Parameter(int data) {
		if (!seq) return;
		switch (seq--) {
			case 3:
				height25 = (data & 0x1f) < 9;
				break;
			case 1:
				colorMode = (data & 0x40) != 0;
				break;
			case 7:
				x = data;
				break;
			case 6:
				y = data;
				seq = 0;
				break;
		}
	}
	int seq, x, y;
	bool width80, height25, colorMode;
};

struct Document;

struct PC8001 : Z80 {
	PC8001(Document *_doc) : doc(_doc), pch(0), vrtc(0) {}
	void SetPC(int _pc) { pc = _pc; }
	bool Extender(uint16_t pc);
	int32_t input(uint16_t adr) {
		int32_t r = 0;
		switch (adr & 0xf0) {
			case 0:
				r = key.d[adr & 0xf];
				break;
			case 0x20:
				r = adr & 1 ? 0xe1 : 0;
				break;
			case 0x40:
				r = (vrtc++ & 0x20) | rtc.data << 4 | 0xa;
				break;
			case 0x80: // custom
				r = (key.d[1] & 0x80) | (key.d[0] >> 1 & 0x20) | (key.d[0] >> 1 & 8) |
				(key.d[0] << 4 & 0x40) | (key.d[1] << 4 & 0x10) | (key.d[9] >> 6 & 1) | 6;
				break;
		}
		return r;
	}
	void output(uint16_t adr, uint8_t data) {
		switch (adr & 0xf0) {
			case 0x10:
				pch = data;
				break;
			case 0x30:
				crtc.width80 = data & 1;
				break;
			case 0x40:
				if (data & 2) rtc.Strobe(pch);
				else if (data & 4) rtc.Shift(pch);
				break;
			case 0x50:
				if (adr & 1) crtc.Command(data);
				else crtc.Parameter(data);
				break;
		}
	}
	int pch, vrtc;
	Key key;
	RealTimeClock rtc;
	CRTController crtc;
	Document *doc;
};

// -------------------------

static int readFile(const char *path, uint8_t *&buf, size_t *size_p = NULL) {
	FILE *fi = fopen(path, "rb");
	if (!fi) return 1;
	fseek(fi, 0, SEEK_END);
	size_t size = ftell(fi);
	rewind(fi);
	if (buf == NULL) buf = new uint8_t[size];
	fread(buf, 1, size, fi);
	fclose(fi);
	if (size_p != NULL) *size_p = size;
	return 0;
}

struct Document {
	struct Header {
		uint16_t jmp, lenb, stm, lenm, cmt_check, pad1, pad2, pad3;
	};
	Document(): pc(NULL), data(NULL), datasize(0) {
		static uint8_t romadd[] = {
			0xff, 0xff, 0xc3, 0x00, 0x00, 0xcd, 0x86, 0x5b,
			0xe7, 0x21, 0x22, 0x60, 0xcd, 0xed, 0x52, 0xc3,
			0x81, 0x00, 0xcd, 0x76, 0x3d, 0x23, 0x22, 0xa0,
			0xef, 0x21, 0x29, 0x60, 0xcd, 0xed, 0x52, 0xc3,
			0x8e, 0x1f, 0x73, 0x61, 0x76, 0x65, 0x64, 0x2e,
			0x00, 0x6c, 0x6f, 0x61, 0x64, 0x65, 0x64, 0x2e,
			0x0d, 0x0a, 0x00
		};
		m = new uint8_t[0x10000];
		int i;
		for (i = 0x6000; i < 0x6000 + sizeof(romadd); i++) m[i] = romadd[i - 0x6000];
		while (i < 0x8000) m[i++] = 0xff;
		while (i < 0x10000) {
			while (!(i & 0x40)) m[i++] = 0xff;
			while (i & 0x40) m[i++] = 0;
		}
	}
	~Document() {
		delete[] m;
		delete[] data;
	}
	int Init(PC8001 *_pc, const char *qs) {
		pc = _pc;
		int result = readFile("pc8001.rom", m);
		if (result) return result;
		if (strlen(qs) > 2) {
			char path[256];
			sprintf(path, "program/%.255s", &qs[2]);
			result = readFile(path, data, &datasize);
			if (result) return result;
			m[KEYIN_HOOK] = 0xcd;
			m[LOAD_HOOK] = EXTENDER_CODE;
		}
		pc->SetMemoryPtr(m);
		return 0;
	}
	void callbackLoad() {
		Header head = *(Header *)data;
		head.jmp = ex16(head.jmp);
		head.lenb = ex16(head.lenb);
		head.stm = ex16(head.stm);
		head.lenm = ex16(head.lenm);
		if (head.cmt_check == 0xd3d3) {
			uint8_t *p, *dp, *start = data + sizeof(Header), *lim = data + datasize;
			for (p = start; p[0] || p[1] || p[2] || p[3] || p[4] || p[5] || p[6] || p[7] || p[8] || p[9];
				 p++)
				;
			memcpy(&m[StartBASIC()], start, p - start + 3);
			while (p < lim && *p++ != 0x3a)
				;
			if (p < lim) {
				dp = &m[p[0] << 8 | p[1]];
				p += 3;
				while (p < lim && *p == 0x3a) {
					p++;
					int n = *p++;
					while (n--) *dp++ = *p++;
					p++;
				}
			}
		}
		else {
			if (head.lenm) memcpy(&m[head.stm], &data[head.lenb + sizeof(Header)], head.lenm);
			if (head.lenb) memcpy(&m[StartBASIC()], &data[sizeof(Header)], head.lenb);
		}
		m[LOAD_HOOK] = 0xcd;
		pc->SetPC(head.lenb ? 0x6012 : head.jmp);
	}
	PC8001 *pc;
	uint8_t *m, *data;
	size_t datasize;
};

bool PC8001::Extender(uint16_t pc) {
	bool f = pc == 0x20;
	if (f) doc->callbackLoad();
	return f;
}

// -------------------------

static void saveKeyData(const char *path, const char *qs) {
	if (strlen(qs) > 2) {
		FILE *fi = fopen(path, "a+b");
		if (fi) {
			uint16_t data = atoi(&qs[2]);
			fwrite(&data, sizeof(data), 1, fi);
			fclose(fi);
		}
	}
}

static void processKeyData(const char *path, PC8001 &pc, int &offset) {
	static uint8_t t[] = {
		0000, 0000, 0000, 0000, 0000, 0000, 0000, 0000,
		0103, 0000, 0000, 0000, 0100, 0017, 0000, 0000,
		0106, 0107, 0105, 0110, 0000, 0000, 0000, 0000, // 16
		0000, 0000, 0000, 0117, 0000, 0000, 0000, 0000,
		0116, 0000, 0000, 0000, 0000, 0000, 0101, 0102, // 32
		0000, 0000, 0000, 0000, 0000, 0000, 0103, 0000,
		0060, 0061, 0062, 0063, 0064, 0065, 0066, 0067, // 48
		0070, 0071, 0000, 0000, 0000, 0000, 0000, 0000,
		0000, 0021, 0022, 0023, 0024, 0025, 0026, 0027, // 64
		0030, 0031, 0032, 0033, 0034, 0035, 0036, 0037,
		0040, 0041, 0042, 0043, 0044, 0045, 0046, 0047, // 80
		0050, 0051, 0052, 0104, 0000, 0104, 0000, 0000,
		0000, 0001, 0002, 0003, 0004, 0005, 0006, 0007, // 96
		0010, 0011, 0012, 0013, 0000, 0000, 0016, 0000,
		0111, 0112, 0113, 0114, 0115, 0000, 0000, 0000, // 112
		0000, 0000, 0000, 0000, 0110, 0110, 0110, 0000,
		0000, 0000, 0000, 0000, 0000, 0000, 0000, 0000, // 128
		0000, 0000, 0000, 0000, 0000, 0000, 0000, 0000,
		0000, 0000, 0000, 0000, 0000, 0000, 0000, 0000, // 144
		0000, 0000, 0000, 0000, 0000, 0000, 0000, 0000,
		0000, 0000, 0000, 0000, 0000, 0000, 0000, 0000, // 160
		0000, 0000, 0000, 0000, 0000, 0000, 0000, 0000,
		0000, 0000, 0000, 0000, 0000, 0000, 0000, 0000, // 176
		0000, 0000, 0073, 0056, 0074, 0057, 0075, 0076,
		0000, 0000, 0000, 0000, 0000, 0000, 0000, 0000, // 192
		0000, 0000, 0000, 0000, 0000, 0000, 0000, 0000,
		0000, 0000, 0000, 0000, 0000, 0000, 0000, 0000, // 208
		0000, 0000, 0000, 0020, 0055, 0053, 0072, 0000,
		0000, 0000, 0000, 0000, 0000, 0000, 0000, 0000, // 224
		0000, 0000, 0000, 0000, 0000, 0000, 0000, 0000,
		0000, 0000, 0000, 0000, 0000, 0000, 0000, 0000,
		0000, 0000, 0000, 0000, 0000, 0000, 0000, 0000,
	};
	FILE *fi = fopen(path, "rb");
	if (fi) {
		fseek(fi, offset, SEEK_SET);
		uint16_t data;
		if (fread(&data, sizeof(data), 1, fi) == 1) {
			offset += sizeof(data);
			if (data) pc.key.Changed(t[data & 0xff], !(data & 0x100));
		}
		fclose(fi);
	}
}

struct Buffer {
	static const int BUFSIZE = 4096, VRAMSIZE = 3000;
	struct Header {
		Header() : 
			sync0(0xff), sync1(0xff), width80(0), height25(0), 
			x(0), y(0), colorMode(0), blink(0), id(0) {}
		Header(CRTController &crtc, int _blink) : 
			sync0(0xff), sync1(0xff), 
			width80(crtc.width80), height25(crtc.height25), 
			x(crtc.x), y(crtc.y), colorMode(crtc.colorMode), 
			blink(_blink), id(getpid()) {}
		uint8_t sync0, sync1, width80, height25, x, y, colorMode, blink;
		uint16_t id;
	};
	Buffer() {
		buf = new uint8_t[BUFSIZE];
		lastvram = new uint8_t[VRAMSIZE];
		memset(lastvram, 0, VRAMSIZE);
	}
	~Buffer() {
		delete[] buf;
		delete[] lastvram;
	}
	int Send(Header *h, uint8_t *vram) {
		memmove(buf, h, sizeof(Header));
		uint8_t *sp = vram, *lim = vram + VRAMSIZE;
		uint8_t *lp = lastvram;
		uint8_t *dp = &buf[sizeof(Header)];
		while (sp < lim) {
			int i, j;
			for (i = 0; &sp[i] < lim && sp[i] == lp[i]; i++)
				;
			*dp++ = i;
			*dp++ = i >> 8;
			if ((sp += i) >= lim) break;
			lp += i;
			for (i = j = 0; &sp[i] < lim && (sp[i] == lp[i] ? ++j <= 4 : !(j = 0)); i++)
				;
			*dp++ = i;
			*dp++ = i >> 8;
			for (j = 0; j < i; j++) *dp++ = sp[j];
			sp += i;
			lp += i;
		}
		int result = 0;
		if (memcmp(h, &lastHeader, sizeof(Header)) || dp > &buf[sizeof(Header) + 2]) {
			lastHeader = *h;
			memmove(lastvram, vram, VRAMSIZE);
			result = !fwrite(buf, dp - buf, 1, stdout);
			fflush(stdout);
		}
		return result;
	}
	Header lastHeader;
	uint8_t *buf, *lastvram;
};

#define FPS		60

int main() {
	printf("Content-type: text/html; charset=utf-8\n\n");
	char keyFilePath[256];
	const char *qs = getenv("QUERY_STRING");
	if (*qs == 'k') {
		const char *p = strstr(qs, "&id=");
		if (p == NULL) return 1;
		sprintf(keyFilePath, "/tmp/pckey%.255s", &p[4]);
		saveKeyData(keyFilePath, qs);
		return 0;
	}
	sprintf(keyFilePath, "/tmp/pckey%d", getpid());
	Document doc;
	PC8001 pc(&doc);
	int result = doc.Init(&pc, qs);
	if (result) return result;
	unlink(keyFilePath);
	int offset = 0, blink = 0;
	Buffer b;
	do {
		timeval tvs, tve;
		gettimeofday(&tvs, NULL);
		processKeyData(keyFilePath, pc, offset);
		pc.Execute(380000 / FPS);
		Buffer::Header h(pc.crtc, 15 * ++blink / (FPS << 3));
		b.Send(&h, &doc.m[0xf300]);
		gettimeofday(&tve, NULL);
		int us = 1000000 * (tve.tv_sec - tvs.tv_sec) + tve.tv_usec - tvs.tv_usec;
		us = 1000000 / FPS - us;
		if (us > 0) usleep(us);
	} while (1);
	unlink(keyFilePath);
	return 0;
}
